@@ -143,6 +143,8 @@ class LuaScripts(object):
                         redis.call('ZREMRANGEBYRANK', el, 0, len_stitch - max_len_stitch_objects - 1) 
                     end
                     redis.call('EXPIRE', el, {user_ttl_with_margin})
+                    table.insert(stitch_keys_with_ts, timestamp_now)
+                    table.insert(stitch_keys_with_ts, el)
                 end
                 redis.call('ZADD', KEYS[7], unpack(stitch_keys_with_ts))
                 redis.call('ZADD', KEYS[9], unpack(stitch_keys_with_ts)) -- this one does not expire
@@ -631,4 +633,38 @@ class LuaScripts(object):
                     end
                 end
                 return result
-            '''
\ No newline at end of file
+            '''
+
+    @classmethod
+    def remove_stitch_keys_from_pool(cls):
+        '''
+        KEYS[1]: stitch pool key
+        KEYS[2]: all stitch pools key
+        KEYS[3]: recently checked db key
+        ARGV: list of stitch keys
+        '''
+
+        return '''
+            redis.call('ZREM', KEYS[1], unpack(ARGV))
+            if redis.call('EXISTS', KEYS[1]) == 0 then
+                redis.call('ZREM', KEYS[2], KEYS[1])
+            end
+            redis.call('DEL', KEYS[3])
+        '''
+
+    @classmethod
+    def remove_user_ids_from_stitch(cls):
+        '''
+        KEYS[1]: stitch key
+        KEYS[2]: all stitch keys key
+        KEYS[3]: recently checked db key
+        ARGV: list of user_ids
+        '''
+
+        return '''
+            redis.call('ZREM', KEYS[1], unpack(ARGV))
+            if redis.call('EXISTS', KEYS[1]) == 0 then
+                redis.call('ZREM', KEYS[2], KEYS[1])
+            end
+            redis.call('DEL', KEYS[3])
+        '''
\ No newline at end of file
