@@ -26,14 +26,16 @@ class StitchPool(object):
     def __init__(self,
                  user_id,
                  stitch_keys=None,
+                 timestamps=None,
                  recently_checked_db=None):
 
         self.user_id = user_id
         self.stitch_keys = set()
+        self.stitch_keys_with_ts = {}
         self._recently_checked_db = recently_checked_db
 
         if stitch_keys:
-            self.add_stitch_keys(stitch_keys)
+            self.add_stitch_keys(stitch_keys, timestamps=timestamps)
 
         if self.domain is None:
             raise ValueError('domain should not be None')
@@ -42,27 +44,50 @@ class StitchPool(object):
             raise ValueError('redis_client should not be None')
 
     def add_stitch_keys(self,
-                        stitch_keys):
+                        stitch_keys,
+                        timestamps=None):
 
         assert isinstance(stitch_keys, (list, set)), 'stitch_keys needs to be a list or set'
+
+        if timestamps:
+            assert isinstance(timestamps, list), 'timestamps should be a list'
+            assert len(timestamps) == len(stitch_keys), 'length of timestamps should be equal to length of stitch_keys'
+            assert all(isinstance(ts, (float, int)) for ts in timestamps), 'all timestamps should be floats or integers'
+        else:
+            now = pytz.utc.localize(datetime.datetime.utcnow()).timestamp()
+            timestamps = [now for _ in stitch_keys]
+
+        self.stitch_keys_with_ts.update({stitch_key: max(ts, self.stitch_keys_with_ts.get(stitch_key, 0)) for stitch_key, ts in
+                                      dict(zip(stitch_keys, timestamps)).items()})
         self.stitch_keys = self.stitch_keys.union(set(stitch_keys))
 
-    def delete_from_cache(self):
-        self.redis_client.delete(self.key)
+    def offload(self):
+        self.get(force_db=True)
+        self.to_db()
+        self.redis_client.remove_stitch_keys_from_pool(
+            keys=[
+                self.key,
+                self.all_stitchpools_key,
+                self.recently_checked_db_key
+            ],
+            args=self.stitch_keys
+        )
+
 
     def to_cache(self):
 
-        now = pytz.utc.localize(datetime.datetime.utcnow()).timestamp()
         pipe = self.redis_client.pipeline()
-        pipe.zadd(self.key, {s: now for s in self.stitch_keys})
+        pipe.zadd(self.key, self.stitch_keys_with_ts)
         pipe.expire(self.key, self.user_ttl + self.user_data_ttl_margin)
         pipe.execute()
 
-    def get(self):
+    def get(self, force_db=False):
 
-        stitch_keys = [s.decode() for s in self.redis_client.zrangebyscore(self.key, min='-inf', max='+inf')]
-        self.add_stitch_keys(stitch_keys)
-        if not self.recently_checked_db:
+        result = self.redis_client.zrangebyscore(self.key, min='-inf', max='+inf', withscores=True)
+        stitch_keys = [s[0].decode() for s in result]
+        timestamps = [s[1] for s in result]
+        self.add_stitch_keys(stitch_keys, timestamps=timestamps)
+        if not self.recently_checked_db or force_db:
             self.complement_from_db()
 
     def complement_from_db(self):
@@ -71,7 +96,8 @@ class StitchPool(object):
         try:
             stitch_pool_model = model.objects.get(user_id=self.user_id)
             stitches_keys = [str(s) for s in stitch_pool_model.stitch_keys]
-            self.add_stitch_keys(stitches_keys)
+            timestamps = [float(ts) for ts in stitch_pool_model.timestamps]
+            self.add_stitch_keys(stitches_keys, timestamps=timestamps)
             if len(stitches_keys) > 0:
                 self.to_cache()
             self.set_recent_db_check()
@@ -79,6 +105,59 @@ class StitchPool(object):
             self.set_recent_db_check()
             return []
 
+    @classmethod
+    def get_expired(cls,
+                    num=300,
+                    start=0,
+                    dev=False):
+
+        now = pytz.utc.localize(datetime.datetime.utcnow()).timestamp()
+        expired_stitchpools = cls.redis_client.zrangebyscore(cls._all_stitchpools_key(),
+                                                             min='-inf',
+                                                             max='+inf' if dev else (now - cls.user_ttl),
+                                                             start=start,
+                                                             num=num,
+                                                             withscores=True)
+
+        return [cls.from_key(stitchpool_key.decode()) for stitchpool_key, timestamp in expired_stitchpools]
+
+    def to_db(self):
+
+        # first call complement_from_db!
+
+        if len(self.stitch_keys_with_ts) == 0:
+            raise Exception('Cannot store empty stitch_keys')
+        max_tries = 3
+        _tmp = list(self.stitch_keys_with_ts.items())
+        stitch_keys = [s[0] for s in _tmp]
+        timestamps = [s[1] for s in _tmp]
+
+        for i in range(max_tries):
+            try:
+                self.usable_django_model().objects.update_or_create(
+                    key=self.key,
+                    defaults={'stitch_keys': stitch_keys, 'timestamps': timestamps})
+                return
+            except Exception as e:
+                if 'OperationalError' in e.__class__.__name__:
+                    error = e
+                    time.sleep(1 + random.random() / 10)
+            raise error
+
+    def remove_stitch_keys(self,
+                           stitch_keys,
+                           remove_from_stitches=False):
+
+        assert isinstance(stitch_keys, (list, set)), 'stitch_keys needs to be a list or set'
+        pipe = self.redis_client.pipeline()
+        pipe.zrem(self.key, *stitch_keys)
+        if remove_from_stitches:
+            for stitch_key in stitch_keys:
+                pipe.zrem(stitch_key, self.user_id)
+        pipe.execute()
+        self.stitch_keys = self.stitch_keys - set(stitch_keys)
+        self.stitch_keys_with_ts = {k: v for k, v in self.stitch_keys_with_ts if not k in stitch_keys}
+
     @classmethod
     def usable_django_model(cls):
         if cls.django_stitch_pool_model:
@@ -105,6 +184,18 @@ class StitchPool(object):
         # this is also defined on the profile class in python-grains!
         return f'st:u:p:{cls.domain}:{user_id}'
 
+    @classmethod
+    def parse_key(cls, key):
+        split_str = key.split(':')
+        user_id = split_str[-1]
+        domain = ':'.join(split_str[3:-1])
+        return domain, user_id
+
+    @classmethod
+    def from_key(cls, key):
+        domain, user_id = cls.parse_key(key)
+        return cls(user_id=user_id)
+
     @property
     def recently_checked_db_key(self):
         return self._recently_checked_db_key(user_id=self.user_id)
@@ -178,29 +269,58 @@ class Stitch(object):
     def delete_from_cache(self):
         self.redis_client.delete(self.key)
 
+    def offload(self):
+        self.get(force_db=True)
+        self.to_db()
+        self.redis_client.remove_user_ids_from_stitch(
+            keys=[
+                self.key,
+                self.all_stitches_key,
+                self.recently_checked_db_key
+            ],
+            args=self.user_ids
+        )
+
     def to_cache(self):
         pipe = self.redis_client.pipeline()
         pipe.zadd(self.key, self.user_ids_with_ts)
         pipe.expire(self.key, self.user_ttl + self.user_data_ttl_margin)
         pipe.execute()
 
-    def get(self):
+    def get(self, force_db=False):
         result = self.redis_client.zrangebyscore(self.key, min='-inf', max='+inf', withscores=True)
         user_ids = [u[0].decode() for u in result]
         timestamps = [u[1] for u in result]
         self.add_user_ids(user_ids, timestamps=timestamps)
-        if not self.recently_checked_db:
+        if not self.recently_checked_db or force_db:
             self.complement_from_db()
 
+    @classmethod
+    def get_expired(cls,
+                    num=300,
+                    start=0,
+                    dev=False):
+
+        now = pytz.utc.localize(datetime.datetime.utcnow()).timestamp()
+        expired_stitches = cls.redis_client.zrangebyscore(cls._all_stitches_key(),
+                                                          min='-inf',
+                                                          max='+inf' if dev else (now - cls.user_ttl),
+                                                          start=start,
+                                                          num=num,
+                                                          withscores=True)
+
+        return [cls.from_key(stitch_key.decode()) for stitch_key, timestamp in expired_stitches]
+
     def remove_user_ids(self,
-                        user_ids):
+                        user_ids,
+                        remove_from_pools=False):
 
         assert isinstance(user_ids, (list, set)), 'user_ids needs to be a list or set'
         pipe = self.redis_client.pipeline()
         pipe.zrem(self.key, *user_ids)
-        for user_id in user_ids:
-            pipe.zrem(self.stitch_pool_key(user_id=user_id), self.key)
-        # pipe.zrem(self.all_stitchpools_key(), )
+        if remove_from_pools:
+            for user_id in user_ids:
+                pipe.zrem(self.stitch_pool_key(user_id=user_id), self.key)
         pipe.execute()
         self.user_ids = self.user_ids - set(user_ids)
         self.user_ids_with_ts = {k: v for k, v in self.user_ids_with_ts if not k in user_ids}
@@ -319,6 +439,11 @@ class Stitch(object):
             else:
                 return cls.django_stitch_model
 
+    @classmethod
+    def from_key(cls, key):
+        type, value, domain = cls.parse_key(key)
+        return cls(value=value, type=type)
+
     @classmethod
     def parse_key(cls, key):
 
